
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/


#include "EvoMod.h"

EvoMod::EvoMod(Node * const & n, vector<double>* const  & mup, Matrix * const  & mpp,  Matrix * const & mur,  Matrix * const & mpr) {
	loadMatrices(mup,mpp,mur,mpr);
	mur->diagonalEintraegeAngleichen();
	mpr->diagonalEintraegeAngleichen();
	calculateTransitionMatrices(n, mur, mpr);
}

EvoMod::~EvoMod() {

}
void EvoMod::loadMatrices(vector<double>* const  & mup, Matrix * const  & mpp,  Matrix * const & mur,  Matrix * const & mpr){
	ifstream file;
	string buf;
	file.open("input_evolution_model");
	if (!file.is_open()) {
		throw "Error loading File (input_evolution_model)";
	}
	int n=0;
	int m=0;
	int matrix=0;
	while (std::getline(file, buf)){
		if(!buf.empty() && buf.at(0)=='|'){
			matrix++;
			n=0;
			m=0;
		}
		else{
			switch(matrix){
				case 0:
					if (!buf.empty())
						mup->push_back(atof(buf.data()));
					break;
				case 1:
					if (!buf.empty()){
						mpp->getTable()[n][m]=atof(buf.data());
						m++;
					}
					else{
						n++;
						m=0;
					}
					break;
				case 2:
					if (!buf.empty()){
						mur->getTable()[n][m]=atof(buf.data());
						m++;
					}
					else{
						n++;
						m=0;
					}
					break;
				case 3:
					if (!buf.empty()){
						mpr->getTable()[n][m]=atof(buf.data());
						m++;
					}
					else{
						n++;
						m=0;
					}
					break;
			}
		}
	}
	file.close();
}
void EvoMod::matrixMulti(Matrix * const & multi, Matrix * const & ma, long double d){
	unsigned int n = ma->getM();
		for(unsigned int i=0; i < n; i++){
			for(unsigned int j=0; j < n; j++){
				multi->getTable()[i][j]=ma->getTable()[i][j]*d;
			}
		}
}
void EvoMod::matrixSquare(Matrix  * const & ma) {
	unsigned int n = ma->getM();
	Matrix * help= new Matrix(n);
	unsigned int i,j,k;
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			help->getTable()[i][j] = 0;
			for (k = 0; k < n; k++) {
				help->getTable()[i][j] += ma->getTable()[i][k] * ma->getTable()[k][j];
			}
		}
	}
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			ma->getTable()[i][j]=help->getTable()[i][j];
		}
	}
	delete(help);
}
void EvoMod::matrixAdition(Matrix * const & ma, Matrix * const & mb) {
	unsigned int n = ma->getM();
	for (unsigned int i = 0; i < n; i++) {
		for (unsigned int j = 0; j < n; j++) {
			ma->getTable()[i][j] = ma->getTable()[i][j] + mb->getTable()[i][j];
		}
	}
}
void EvoMod::transitionDynamicsPair(Matrix * const & pt,Matrix * const & pairedR, double const dist){
	long double k= 30;
	long double n= powl(2,k);
	Matrix * I= new Matrix(16);
	for(unsigned int i=0; i < I->getM(); i++){
		for(unsigned int j=0; j < I->getM(); j++){
			if(i==j){
				I->getTable()[i][j]=1;
			}
			else{
				I->getTable()[i][j]=0;
			}
		}
	}
	long double tn = dist/n;
	matrixMulti(pt, pairedR, tn);
	matrixAdition(pt,I);
	delete(I);
	for(unsigned int i=0;i<k;i++){
		matrixSquare(pt);
	}
 }
void EvoMod::transitionDynamics(Matrix * const & pt,  Matrix * const & unpairedR, double const dist){
	long double k= 30;
	long double n= powl(2,k);
	Matrix * I= new Matrix(4);
	for(unsigned int i=0; i < I->getM(); i++){
		for(unsigned int j=0; j < I->getM(); j++){
			if(i==j){
				I->getTable()[i][j]=1;
			}
			else{
				I->getTable()[i][j]=0;
			}
		}
	}
	long double tn = dist/n;
	matrixMulti(pt, unpairedR, tn);
	matrixAdition(pt,I);
	delete(I);
	for(unsigned int i=0;i<k;i++){
		matrixSquare(pt);
	}
 }

void EvoMod::calculateTransitionMatricesReal(Node * const & r, Matrix * const & unpairedR, Matrix * const & pairedR){
	 if (!r->isLeaf()) {
		 transitionDynamics(r->getPt(),unpairedR, r->getDistance());
		 transitionDynamicsPair(r->getPtPaired(),pairedR, r->getDistance());
		 calculateTransitionMatricesReal( r->getLeftChild(),unpairedR, pairedR);
		 calculateTransitionMatricesReal( r->getRightChild(),unpairedR, pairedR);
	   }
	   else{
		   transitionDynamics(r->getPt(),unpairedR, r->getDistance());
		transitionDynamicsPair(r->getPtPaired(),pairedR, r->getDistance());
	   }
}
void EvoMod::calculateTransitionMatrices(Node * const & r, Matrix * const & unpairedR, Matrix * const & pairedR){
	if(r->amount()<3){
		cerr<<"Baum ist zu klein"<<endl;
		exit(1);
	}
	else{
		calculateTransitionMatricesReal(r->getLeftChild(),unpairedR, pairedR);
		calculateTransitionMatricesReal(r->getRightChild(), unpairedR,  pairedR);
	}
}
