
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/


#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <ctime>
#include <string>
#include <cmath>
#include <cstdlib>
#include <limits>

using namespace std;

/*  Grammar:
 *
 *  S -> LS | x | axFyb | Q
 *  F -> xLSy | xFy  
 *  L -> x | axFyb | Q
 */
#ifndef HSP_H_
#define HSP_H_
class Hsp {
public:
    int x,y; //Koordinaten des Startpunkts
    int len; //Laenge
    double score, wscore;
    // score und gewichteter score
    double logemisswkeit;
    // Wahrscheinlichkeit der Basentypen falls stem
    Hsp(int x1, int y1, int len1, double score1, double wscore1, double lewk):
    	x(x1),y(y1),len(len1),
    	score(score1),wscore(wscore1),logemisswkeit(lewk)
    {
    }
};
#endif /* HSP_H_ */

#include "Node.h"
#ifndef SAMPLER_H_
#define SAMPLER_H_
class Sampler {

private :
	map <string, double>  * columnLikelihood;
	map <string, double>  * columnPairLikelihood;
    //referenz auf pointer der orginial datenstruktur welcher ursprünglich const ist
	//jedoch auf grund des zuweisungsoperators von Sampler nicht const sein kann ausser ich ueberlade diesen zuweisungsoperator...
    vector <string > * originalAlignment;
    //copie auf die datenstruktur;
    vector<string> mAlignment;
    // Der Vektor qregion enthaelt abwechselnd Anfang und Ende von Q-Regionen
    vector<size_t> qregion;
	vector<Hsp> diag;
	double totalwsc;

    map<char,vector<vector<double> > > alpha;
    /* Jeder Nichtterminalen X der Grammatik ordnet die map
     * alpha einen vector<vector<double> > zu, so dass
     * alpha[X][i][j] der log der Wahrscheinlichkeit ist, dass 
     * aus einer Variablen X letzlich die Terminalenfolge 
     * sequenz[i],sequenz[i+1],...,sequenz[j] wird. 
     *
     * Zusatz: in wahrheit bis [i+j]
     */
    map<char,vector<vector<double> > > beta;
    /* Jeder Nichtterminalen X der Grammatik ordnet die map
     * beta einen vector<vector<double> > zu, so dass
     * beta[X][i][j] der log der Wahrscheinlichkeit ist, dass 
     * am Ende etwas entsteht, das mit
     * sequenz[0],...,sequenz[i-1] beginnt, mit 
     * sequenz[j+1],...,sequenz[n-1] endet und dazwischen 
     * etwas stehen hat, das von einem X erzeugt wurde.
     */
    vector<vector<double> > posteriwpaar;
    /* posteriwpaar[i][j] ist die a posteriori Wahrscheinlichkeit, dass Positionen
     * [i] und [j] in einem Stem miteinander gepaart sind.
     */

    void inside_algo();
    // berechnet alpha mit inside-Algo
    // berechnet alpha mit inside-Algo; schnellere Version
    void inside_algo_original();
    void outside_algo();
    // berechnet beta
    double StoLS, Stox, StoaxFyb, StoQ, 
	FtoxFy, FtoxLSy, Ltox, LtoaxFyb, LtoQ,
	qparam;
    double logStoLS, logStox, logStoaxFyb, logStoQ, 
	logFtoxFy, logFtoxLSy, logLtox, logLtoaxFyb, logLtoQ,
	logqparam;
    void lieswkeiten(const char * dateiname); 
    inline double addlog(const double  a, const double &b) {
		// liefert log(exp(a)+exp(b))
		if (a>b) return a+log(1+exp(b-a));
		else return b+log(1+exp(a-b));
    };
    // Wahrscheinlichkeit des Basenpaars xy in Stem
    string sample(int i, int j, char X);
    // sample ausgehend von non-Terminal X eine Faltung fuer 
    // sequenz[i],...,sequenz[j], wobei Q als Q ausgegeben wird.
    void alignment();
    void alignmentNew();
    void findPseudoKnots();
    // Findet und scort alle HSPs der Sequenz mit ihrem Komplement
    // und weist ihnen einen gewichteten Score zu.
public :
    Sampler(map <string, double>  * const & columnLikelihood, map <string, double>  * const & columnPairLikelihood,
    		 vector<string > * const & rpAlignment, const char * wkeitendatei,const vector<size_t> & qreg);
    // Der Vektor qregion enthaelt abwechselnd Anfang und Ende von Q-Regionen
    Sampler( map <string, double>  * const & columnLikelihood,map <string, double>  * const & columnPairLikelihood,
    		vector<string > * const & rpAlignment, const char * const wkeitendatei);
    inline void loesche_maps() {
    	alpha.clear();
    	beta.clear();
    	diag.clear();
    }

    void set_qreg(const vector<size_t> & qreg);
    string sample() {
    	return sample(0,mAlignment.size()-1,'S');
    }
    // Sampelt eine Sekundaerstruktur ohne Pseudoknoten
    inline double logwkeit() {
    	return alpha['S'][0][mAlignment.size()-1];
    }
    double get_posteriwpaar(size_t i,size_t j) {
		if (posteriwpaar.size()==0)
			outside_algo();
		if ( i>mAlignment.size() || j> mAlignment.size())
			return 0;
		return posteriwpaar[i][j];
    }
    /*
    void posteriausgabe(char * datei) {
		if (posteriwpaar.size()==0)
			outside_algo();
		ofstream dat(datei);
		for (size_t i=0; i<sequenz.size(); ++i){
			for (size_t j=0; j<sequenz.size(); ++j){
				dat << posteriwpaar[i][j] << endl;
			}
		}
		dat.close();
    }
	*/
    Hsp randhsp();
    // Gibt ein zufaelliges Element von diag aus, gemaess dem wscore.
    double get_totalwsc() {
    	return totalwsc;
    }
    vector<Hsp> get_diag(){
		if (diag.size()==0)
			alignment();
		return diag;
    }



    inline map <string, double> *  get_columnLikelihood() const {
       	return  columnLikelihood;
    }
    inline map <string, double> *  get_columnPairLikelihood() const {
		return  columnPairLikelihood;
	}
    inline vector <string > *  get_OriginalAlignment() const {
		return  originalAlignment;
	}


    inline double getqparam() const {return qparam;}
    inline vector<size_t> get_qregion() const {return qregion;} 
    inline double get_logStoLS() const {return logStoLS;}
    inline double get_logStox() const {return logStox;}  
    inline double get_logStoaxFyb() const {return logStoaxFyb;} 
    inline double get_logStoQ() const {return logStoQ;} 
    inline double get_logFtoxFy() const {return logFtoxFy;} 
    inline double get_logFtoxLSy() const {return logFtoxLSy;} 
    inline double get_logLtox() const {return logLtox;}
    inline double get_logLtoaxFyb() const {return logLtoaxFyb;} 
    inline double get_logLtoQ() const {return logLtoQ;}
    inline double get_logqparam() const {return logqparam;}

};
#endif /* SAMPLER_H_ */


/*
 *
 *                          CYK   CYK   CYK
 */
#ifndef CYK_H_
#define CYK_H_
class Cyk {
private:
	map <string, double>  *  columnLikelihood;
	map <string, double>  *  columnPairLikelihood;
	vector <string > * originalAlignment;
	vector<string> mAlignment;

	vector<size_t> qregion;
	    // Der Vektor qregion enthaelt abwechselnd Anfang und Ende von Q-Regionen
    string faltung;
    // In dem string faltung wird jede Q-Region durch eine 0 ersetzt
    // Darum herum wird die Folge der Zeichen "(",")" und "." ausgegeben,
    // die durch den wahrscheinlichsten Ableitungsbaum induziert wird.
    double logwkeit;
    double  qparam;
    double logStoLS, logStox, logStoaxFyb, logStoQ, 
	logFtoxFy, logFtoxLSy, logLtox, logLtoaxFyb, logLtoQ,
	logqparam;
    void falte();
    void falteNew();
public:
    Cyk(map <string, double>  * const & columnLikelihood, map <string, double>  * const & columnPairLikelihood,
    	vector<string > * const & rpAlignment);
    Cyk(const Sampler & S);
    Cyk(const Cyk & C, const vector<size_t> & qreg);
    void setz(const Sampler & S, const vector<size_t> & qreg);
    inline string get_faltung() const {return faltung;}
    inline double get_logwkeit() const {return logwkeit;}
    inline double get_qparam() const {return qparam;}    

};
#endif /* CYK_H_ */
